/*
 * Copyright (c) 2014 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Tue Feb 18 2014
 */
#include "acoptitrack.h"

#include <sys/select.h>
#include <sys/time.h>
#include <assert.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "optitrack_c_types.h"
#include "codels.h"


/* --- Task publish ----------------------------------------------------- */


/* --- Activity connect ------------------------------------------------- */

static void	opti_connect_cleanup(optitrack_server_s **server);

/** Codel opti_connect_start of activity connect.
 *
 * Triggered by optitrack_start.
 * Yields to optitrack_descr.
 * Throws optitrack_e_addrinfo, optitrack_e_network,
 * optitrack_e_nomem, optitrack_e_publish.
 */
genom_event
opti_connect_start(const char host[128], const char host_port[128],
                   const char mcast[128], const char mcast_port[128],
                   optitrack_server_s **server)
{
  optitrack_e_addrinfo_detail d;

  /* private context */
  *server = calloc(1, sizeof(**server));
  if (!*server) return optitrack_e_nomem();

  /* setup network */
  (*server)->natnet = natnet_connect(host, host_port, mcast, mcast_port,
                                     d.what, sizeof(d.what));
  if (!(*server)->natnet) {
    warnx("%s", d.what);
    opti_connect_cleanup(server);
    return optitrack_e_addrinfo(&d);
  }

  return optitrack_descr;
}

/** Codel opti_connect_descr of activity connect.
 *
 * Triggered by optitrack_descr.
 * Yields to optitrack_recv.
 * Throws optitrack_e_addrinfo, optitrack_e_network,
 * optitrack_e_nomem, optitrack_e_publish.
 */
genom_event
opti_connect_descr(optitrack_server_s **server)
{
  optitrack_e_network_detail n;
  int e;

  e = natnet_snd_modeldef((*server)->natnet, n.what, sizeof(n.what));
  if (e) {
    warnx("%s", n.what);
    opti_connect_cleanup(server);
    return optitrack_e_network(&n);
  }

  return optitrack_recv;
}

/** Codel opti_connect_recv of activity connect.
 *
 * Triggered by optitrack_recv.
 * Yields to optitrack_recv, optitrack_command, optitrack_publish, optitrack_stop.
 * Throws optitrack_e_addrinfo, optitrack_e_network,
 * optitrack_e_nomem, optitrack_e_publish.
 */
genom_event
opti_connect_recv(const optitrack_server_s *server)
{
  struct timeval tv = { .tv_sec = 5, .tv_usec = 0 };
  int e;

  e = natnet_poll(server->natnet, &tv);
  if (e == NATNET_EVERR) return optitrack_stop;

  if (e & NATNET_EVCMD) return optitrack_command;
  if (e & NATNET_EVDAT) return optitrack_publish;

  return optitrack_recv;
}

/** Codel opti_connect_command of activity connect.
 *
 * Triggered by optitrack_command.
 * Yields to optitrack_descr, optitrack_recv.
 * Throws optitrack_e_addrinfo, optitrack_e_network,
 * optitrack_e_nomem, optitrack_e_publish.
 */
genom_event
opti_connect_command(optitrack_server_s **server,
                     optitrack_ids_ping_ctx *ping,
                     const optitrack_bodies *bodies)
{
  optitrack_e_network_detail d;
  genom_event e;
  int s;

  s = natnet_cmd_recv((*server)->natnet, d.what, sizeof(d.what));
  switch (s) {
    case NATNET_MSGPING:
      natnet_rcv_ping((*server)->natnet,
                      ping->response, sizeof(ping->response));
      ping->id++;
      break;

    case NATNET_MSGMODELDEF:
      (*server)->descr = natnet_rcv_modeldef((*server)->natnet);
      if ((*server)->descr) {
        e = opti_create_ports((*server)->descr, bodies);
        if (e != genom_ok) return e;
      }
      break;

    case NATNET_MSGERR:
      opti_connect_cleanup(server);
      return optitrack_e_network(&d);
  }

  if (!(*server)->descr) return optitrack_descr;
  return optitrack_recv;
}

/** Codel opti_connect_publish of activity connect.
 *
 * Triggered by optitrack_publish.
 * Yields to optitrack_descr, optitrack_recv.
 * Throws optitrack_e_addrinfo, optitrack_e_network,
 * optitrack_e_nomem, optitrack_e_publish.
 */
genom_event
opti_connect_publish(optitrack_server_s **server,
                     const optitrack_log_s *log,
                     const optitrack_bodies *bodies)
{
  optitrack_e_network_detail d;
  int e;

  e = natnet_data_recv((*server)->natnet, d.what, sizeof(d.what));
  switch (e) {
    case NATNET_MSGFRAMEOFDATA: {
      const struct natnet_frame *f;

      f = natnet_rcv_frameofdata((*server)->natnet);
      if (f)
        opti_update_ports(f, *server, bodies, log?log->f:NULL);
      break;
    }

    case NATNET_MSGERR:
      opti_connect_cleanup(server);
      return optitrack_e_network(&d);
  }

  if (!(*server)->descr) return optitrack_descr;
  return optitrack_recv;
}

/** Codel opti_connect_stop of activity connect.
 *
 * Triggered by optitrack_stop.
 * Yields to optitrack_ether.
 * Throws optitrack_e_addrinfo, optitrack_e_network,
 * optitrack_e_nomem, optitrack_e_publish.
 */
genom_event
opti_connect_stop(optitrack_server_s **server)
{
  opti_connect_cleanup(server);
  return optitrack_ether;
}

/** opti_connect_cleanup
 */
static void
opti_connect_cleanup(optitrack_server_s **server)
{
  if (!server || !*server) return;
  natnet_disconnect((*server)->natnet);
  free(*server);
  *server = NULL;
}
